"""
Author: Miroslav Bártík
Name of the program: Lloydova 15
Version: 1.0
date: 18.11.2024
description: This model serves as a base for all the data (in this case the puzzle)
"""

import random


class Puzzle:

    def __init__(self, n: int, grid):
        """
        Initializes the Puzzle instance.

        :param n: Size of the puzzle grid (n x n).
        :param grid: The starting grid configuration. If None, a solved grid is generated.
        """
        self.n = n
        if grid == None:
            # Generate a solved grid if no grid is provided
            self.grid = self.generate_solved_grid(n)
            self.empty_pos = (n - 1, n - 1)
        else:
            # Use the provided grid and find the empty position
            self.grid = grid
            self.empty_pos = self.find_empty_pos()

    def __lt__(self, other):
        """
       Compares two puzzles based on their Manhattan distance heuristic.

       :param other: Another Puzzle instance.
       :return: True if the current puzzle's Manhattan distance is less than the other's.
       """
        return self.manhattan_distance() < other.manhattan_distance()

    def manhattan_distance(self) -> int:
        """
        Computes the Manhattan distance of the current puzzle state.
        The Manhattan distance is the sum of vertical and horizontal distances
        of each tile from its target position.

        :return: Total Manhattan distance for the grid.
        """
        distance = 0
        for i in range(self.n):
            for j in range(self.n):
                value = self.grid[i][j]
                if value != 0:  # Ignore the empty tile
                    # Calculate target position of the current tile
                    target_x, target_y = divmod(value - 1, self.n)
                    # Add the Manhattan distance for the current tile
                    distance += abs(target_x - i) + abs(target_y - j)
        return distance

    @staticmethod
    def generate_solved_grid(n: int):
        """
        Generates a solved puzzle grid.

        :param n: Size of the grid (n x n).
        :return: Solved n x n grid.
        """
        return [[(i * n + j + 1) % (n * n) for j in range(n)] for i in range(n)]

    def shuffle(self, moves=100):
        """
       Shuffles the puzzle by making random valid moves.

       :param moves: Number of shuffle moves to perform.
       """
        directions = ["up", "down", "left", "right"]
        successful_moves = 0
        while successful_moves < moves:
            # Attempt a random move and count only successful ones
            if self.move(random.choice(directions)):
                successful_moves += 1  # Count only successful moves

    def find_empty_pos(self):
        """
        Locates the position of the empty tile (value 0) in the grid.

        :return: Tuple (x, y) representing the position of the empty tile.
        """
        for x, row in enumerate(self.grid):
            for y, value in enumerate(row):
                if value == 0:
                    return x, y
        raise ValueError("0 not found in the grid.")  # Raise an error if no empty tile is found

    def move(self, direction):
        """
        Moves the empty tile in the specified direction, if valid.

        :param direction: Direction to move ("up", "down", "left", "right").
        :return: True if the move was successful, False otherwise.
        """
        empty_x, empty_y = self.empty_pos

        # Move down
        if direction == "down" and empty_x < self.n - 1:
            self.grid[empty_x][empty_y], self.grid[empty_x + 1][empty_y] = self.grid[empty_x + 1][empty_y], \
                self.grid[empty_x][empty_y]
            self.empty_pos = (empty_x + 1, empty_y)  # Update empty tile position
            return True

        # Move up
        elif direction == "up" and empty_x > 0:
            self.grid[empty_x][empty_y], self.grid[empty_x - 1][empty_y] = self.grid[empty_x - 1][empty_y], \
                self.grid[empty_x][empty_y]
            self.empty_pos = (empty_x - 1, empty_y)  # Update empty tile position
            return True

        # Move right
        elif direction == "right" and empty_y < self.n - 1:
            self.grid[empty_x][empty_y], self.grid[empty_x][empty_y + 1] = self.grid[empty_x][empty_y + 1], \
                self.grid[empty_x][empty_y]
            self.empty_pos = (empty_x, empty_y + 1)  # Update empty tile position
            return True

        # Move left
        elif direction == "left" and empty_y > 0:
            self.grid[empty_x][empty_y], self.grid[empty_x][empty_y - 1] = self.grid[empty_x][empty_y - 1], \
                self.grid[empty_x][empty_y]
            self.empty_pos = (empty_x, empty_y - 1)  # Update empty tile position
            return True
        else:
            # Invalid move
            return False

    def copy(self):
        """
        Creates a deep copy of the current puzzle instance.

        :return: A new Puzzle instance with the same grid and empty position.
        """
        new_puzzle = Puzzle(self.n, self.grid)   # Create a new Puzzle object
        new_puzzle.grid = [row[:] for row in self.grid]   # Deep copy the grid
        new_puzzle.empty_pos = self.empty_pos   # Copy the empty position
        return new_puzzle
