"""
This module @FactBase represents a data manager in the airport knowledge-based system.
It stores and manages data about the airport, aircraft, and flight schedules.
It also provides functionality to generate a random flight table.

Author:
    Václav Hajšman (hajsman2@student-zcu.cz)
Date:
    10.12.2024
Version:
    2.0
"""

from datetime import time
import random

from DataStructures.Aircraft import Aircraft
from DataStructures.Airport import Airport
from DataStructures.Flight import Flight


class Facts:
    """
    Represents the data manager for the airport system.

    Attributes:
        airport (Airport): The airport data.
        aircraftList (list[Aircraft]): A list of Aircraft objects.
        flightTable (list[Flight]): A list of Flight objects.
    """

    def __init__(self):
        """
        Initializes the Facts object with no data.
        """
        self.airport = None
        self.aircraftList = None
        self.flightTable = None

    def setAirportData(self, airport: Airport):
        """
        Sets the airport data.

        Parameters:
            airport (Airport): The airport data to be stored.
        """
        self.airport = airport

    def setAircraftData(self, aircraftList: list[Aircraft]):
        """
        Sets the list of aircraft data.

        Parameters:
            aircraftList (list[Aircraft]): The list of Aircraft objects.
        """
        self.aircraftList = aircraftList

    def setFlightTableData(self, flightTable: list[Flight]):
        """
        Sets the flight table data.

        Parameters:
            flightTable (list[Flight]): The list of Flight objects.
        """
        self.flightTable = flightTable

    def generateFlightTable(self, count: int):
        """
        Generates a random flight table.

        This method creates a specified number of flights by randomly selecting aircraft
        and generating random arrival and departure times. The resulting flight table is
        sorted by arrival time.

        Parameters:
            count (int): The number of flights to generate.
        """
        flightTable = []

        for i in range(count):
            aircraft = random.choice(self.aircraftList)
            time1 = time(random.randint(0, 23), random.randint(0, 59))
            time2 = time(random.randint(0, 23), random.randint(0, 59))

            # Ensure arrival time is earlier than departure time
            if time1 < time2:
                flightTable.append(Flight(aircraft, time1, time2))
            else:
                flightTable.append(Flight(aircraft, time2, time1))

        # Sort flights by arrival time
        sorted_flights = sorted(flightTable, key=lambda f: f.arrival)
        self.flightTable = sorted_flights
