"""
This module defines the class @InferenceMachine, which serves as the central control
for the airport knowledge-based system. It controls the flow of data, calculations,
and explanations related to flight scheduling and gate assignments in an airport.

Author:
    Václav Hajšman (hajsman2@student-zcu.cz)
Date:
    10.12.2024
Version:
    2.0
"""
from ExplanationModule import ExplanationModule
from FactBase import Facts
from KnowledgeBase import Knowledge
from SolutionGenerator import SolutionGenerator
from UserInterface import UserInterface


class InferenceMachine:
    """
    The central controller of the airport knowledge-based system, responsible for
    managing the overall workflow of reading data, performing calculations, and
    providing explanations for flight and gate assignments.

    Attributes:
        knowledge (Knowledge): Holds the knowledge about airport and flight data.
        facts (Facts): Manages gate occupations and handles flight fitting and reallocation.
        userInterface (UserInterface): Handles interaction with the user, including input and output.
        solutionGenerator (SolutionGenerator): Calculates and generates solutions related to flight and gate assignments.
        explanationModule (ExplanationModule): Provides explanations based on the flight data and system rules.
    """

    def __init__(self):
        """
        Initializes the `InferenceMachine` by setting up necessary components.

        Initializes:
            - Knowledge: For the functions which provide the calculations.
            - Facts: For storing user given data.
            - UserInterface: For interacting with the user.
            - SolutionGenerator: For calculating and generating solutions for flight and gate assignments.
            - ExplanationModule: For explaining flight details and decision-making.
        """
        self.knowledge = Knowledge()
        self.facts = Facts()
        self.userInterface = UserInterface()
        self.solutionGenerator = SolutionGenerator()
        self.explanationModule = ExplanationModule()

    def start(self):
        """
        Starts the knowledge-based system by executing a series of steps:
            1. Print program info.
            2. Read configuration files.
            3. Retrieve flight table data.
            4. Initialize gate occupation list.
            5. Perform calculations for flight and gate assignments.
            6. Print the results of the calculations.
            7. Provide explanations for flight details.
        """
        self.validateProgrammeArgs()
        self.printProgrammeInfo()
        self.readConfigFile()
        self.getFlightTable()
        self.printLoadedData()
        self.initOccupationList()
        self.startCalculation()
        self.printOutput()
        self.startExplaining()

    def validateProgrammeArgs(self):
        """
        Validates if the programme arguments are in correct format.
        """
        self.userInterface.validateProgrammeArgs()

    def printProgrammeInfo(self):
        """
        Prints general information about the program to the user via the user interface.
        """
        self.userInterface.printProgrammeInfo()

    def readConfigFile(self):
        """
        Reads the configuration file for airport and aircraft data via the user interface,
        and stores the data in the `Data` object.

        The method retrieves two sets of data:
            - airportData: Data related to the airport (gates, free park count, etc.).
            - aircraftData: Data related to the aircraft (type, gates, etc.).
        """
        airportData, aircraftData = self.userInterface.getConfigFileData()
        self.facts.setAircraftData(aircraftData)
        self.facts.setAirportData(airportData)

    def getFlightTable(self):
        """
        Retrieves the flight table data, either from user input or a predefined set.
        Prompts the user to choose between entering flight data manually or selecting a predefined table.
        """
        userInput = self.userInterface.flightTableChoice()
        if userInput < 1:
            flightData = self.userInterface.getFlightTableData(self.facts.aircraftList)
            self.facts.setFlightTableData(flightData)
        else:
            self.facts.generateFlightTable(userInput)

    def printLoadedData(self):
        """
        Prints the loaded data (airport, aircraft, and flight table) to the user using the user interface.
        """
        self.userInterface.printLoadedData(self.facts)

    def initOccupationList(self):
        """
        Initializes the gate occupation list based on the airport data. This sets up the list of gates
        and their current occupations for the system to manage. Also sets other crucial parameters
        to ensure that the calculation works correctly.
        """
        self.knowledge.setFacts(self.facts)
        self.solutionGenerator.setKnowledge(self.knowledge)
        self.solutionGenerator.setFacts(self.facts)
        self.solutionGenerator.initGateOccupationList(self.facts.airport.gates)

    def startCalculation(self):
        """
        Executes the flight and gate assignment calculations. This involves using the `SolutionGenerator`
        to perform the necessary calculations.
        """
        self.solutionGenerator.calculate()

    def printOutput(self):
        """
        Prints the results of the calculations, including the gate occupation list and any canceled flights.
        This method calls the `UserInterface` to display the output.
        """
        self.userInterface.printOutput(self.solutionGenerator.gateOccupationList, self.solutionGenerator.canceledFlights)

    def startExplaining(self):
        """
        Initiates the explanation process for flight details. The user is prompted to enter flight IDs,
        and the system provides detailed explanations for each flight's gate assignment or cancellation.

        This loop continues until the user inputs `None`.
        """
        self.explanationModule.setKnowledge(self.knowledge)
        self.explanationModule.setFacts(self.facts)

        flightId = self.userInterface.readFlightId()
        while flightId is not None:
            explanations = self.explanationModule.explainFlight(int(flightId), self.solutionGenerator.gateOccupationList, self.solutionGenerator.canceledFlights)
            self.userInterface.printExplanation(explanations)
            flightId = self.userInterface.readFlightId()
