"""
This class @Knowledge represents an object with encapsulates all key methods
which generate the output of the airport knowledge system. It performs the
calculations which determine the occupation of airport gates by flights.

Author:
    Václav Hajšman (hajsman2@student-zcu.cz)
Date:
    10.12.2024
Version:
    2.0
"""
from datetime import time
from DataStructures.Flight import Flight
from DataStructures.Gate import Gate
from FactBase import Facts


class Knowledge:
    """
    Represents the knowledge (functions) for the airport system.

    Attributes:
        facts (Facts): The facts containing all the user given information about the airport.
    """

    def __init__(self):
        """
        Initializes a Data object with default values.
        """
        self.facts = None

    def setFacts(self, facts: Facts):
        """
        Sets the facts for the KnowledgeBase.

        Parameters:
            facts (Facts): The facts containing all the user given information about the airport.
        """
        self.facts = facts

    def addOccupation(self, gate: Gate, flight: Flight, gateOccupationList: dict):
        """
        Adds a flight occupation to the specified gate or free parking space.

        Parameters:
            gate (Gate): The gate or free parking space where the flight is being scheduled.
            flight (Flight): The flight that occupies the gate or parking space.
            gateOccupationList (dict): dictionary with the airport gate occupation data.
        """
        if not gate:
            gateOccupationList["freePark"].append(flight)
        else:
            gateOccupationList[gate.name].append(flight)

    def removeOccupation(self, gate: Gate, arrival: time, departure: time, gateOccupationList: dict):
        """
        Removes a flight occupation from the specified gate or free parking space.

        Parameters:
            gate (Gate): The gate or free parking space where the flight is being removed.
            arrival (time): The arrival time of the flight to be removed.
            departure (time): The departure time of the flight to be removed.
            gateOccupationList (dict): dictionary with the airport gate occupation data.
        """
        if not gate:
            gateOccupationList["freePark"] = [
                flight for flight in gateOccupationList["freePark"]
                if not (flight.arrival < departure and flight.departure > arrival)
            ]
        else:
            gateOccupationList[gate.name] = [
                flight for flight in gateOccupationList[gate.name]
                if not (flight.arrival < departure and flight.departure > arrival)
            ]

    def isOccupied(self, gate: Gate, arrival: time, departure: time, gateOccupationList: dict) -> bool:
        """
        Checks if a gate is occupied by a flight during the given time period.

        Parameters:
            gate (Gate): The gate to check for occupation.
            arrival (time): The arrival time to check for conflicts.
            departure (time): The departure time to check for conflicts.
            gateOccupationList (dict): dictionary with the airport gate occupation data.

        Returns:
            bool: True if the gate is occupied, False otherwise.
        """
        for flight in gateOccupationList[gate.name]:
            if (arrival < flight.departure) and (departure > flight.arrival):
                return True
        return False

    def getFlightsCount(self, gate: Gate, arrival: time, departure: time, gateOccupationList: dict) -> int:
        """
        Returns the count of flights occupying a gate or free parking space during the given time period.

        Parameters:
            gate (Gate): The gate to check the number of flights for.
            arrival (time): The arrival time to check for conflicts.
            departure (time): The departure time to check for conflicts.
            gateOccupationList (dict): dictionary with the airport gate occupation data.

        Returns:
            int: The number of flights occupying the gate or free parking space.
        """
        count = 0
        if not gate:
            for flight in gateOccupationList["freePark"]:
                if (arrival < flight.departure) and (departure > flight.arrival):
                    count += 1
        else:
            for flight in gateOccupationList[gate.name]:
                if (arrival < flight.departure) and (departure > flight.arrival):
                    count += 1

        return count

    def getOccupiedFlight(self, gate: Gate, arrival: time, departure: time, gateOccupationList: dict) -> Flight | None:
        """
        Returns a flight occupying the gate during the given time period, if any.

        Parameters:
            gate (Gate): The gate to check for occupied flights.
            arrival (time): The arrival time to check for conflicts.
            departure (time): The departure time to check for conflicts.
            gateOccupationList (dict): dictionary with the airport gate occupation data.

        Returns:
            Flight | None: The occupied flight, or None if no flight is occupying the gate.
        """
        if self.isOccupied(gate, arrival, departure, gateOccupationList) and self.getFlightsCount(gate, arrival, departure, gateOccupationList) <= 1:
            for flight in gateOccupationList[gate.name]:
                if (arrival < flight.departure) and (departure > flight.arrival):
                    return flight

        return None

    def getOccupiedFlightTotal(self, gate: Gate, arrival: time, departure: time, gateOccupationList: dict) -> list[Flight] | None:
        """
        Returns a list of all flights occupying the gate during the given time period.

        Parameters:
            gate (Gate): The gate to check for occupied flights.
            arrival (time): The arrival time to check for conflicts.
            departure (time): The departure time to check for conflicts.
            gateOccupationList (dict): dictionary with the airport gate occupation data.

        Returns:
            list[Flight] | None: A list of occupied flights, or None if no flights are occupying the gate.
        """
        if self.isOccupied(gate, arrival, departure, gateOccupationList) and self.getFlightsCount(gate, arrival, departure, gateOccupationList) >= 1:
            flights = []
            for flight in gateOccupationList[gate.name]:
                if (arrival < flight.departure) and (departure > flight.arrival):
                    flights.append(flight)
            return flights

        return None

    def reallocateFlight(self, flight: Flight, gateOccupationList: dict) -> bool:
        """
        Attempts to reallocate a flight to a new gate if the current gate is occupied.

        Parameters:
            flight (Flight): The flight to reallocate.
            gateOccupationList (dict): dictionary with the airport gate occupation data.

        Returns:
            bool: True if the flight was successfully reallocated, False otherwise.
        """
        swap = False
        for possibleGate in flight.aircraft.gates:
            if self.getFlightsCount(possibleGate, flight.arrival, flight.departure, gateOccupationList) > 1:
                continue

            oldFlight = self.getOccupiedFlight(possibleGate, flight.arrival, flight.departure, gateOccupationList)
            if oldFlight.aircraft.freePark:
                self.removeOccupation(possibleGate, flight.arrival, flight.departure, gateOccupationList)
                self.addOccupation(possibleGate, flight, gateOccupationList)
                if not self.fitFlight(oldFlight, gateOccupationList):
                    if self.getFlightsCount(None, oldFlight.arrival,
                                            oldFlight.departure,gateOccupationList) < self.facts.airport.freeParkCount:
                        self.addOccupation(None, oldFlight, gateOccupationList)
                swap = True
                break

        return swap

    def fitFlight(self, flight: Flight, gateOccupationList: dict) -> bool:
        """
        Checks if a flight can fit within the available gates.

        Parameters:
            flight (Flight): The flight to check for scheduling conflicts.
            gateOccupationList (dict): dictionary with the airport gate occupation data.

        Returns:
            bool: True if the flight can be scheduled, False otherwise.
        """
        addition = False
        for possibleGate in flight.aircraft.gates:
            if not self.isOccupied(possibleGate, flight.arrival, flight.departure, gateOccupationList):
                self.addOccupation(possibleGate, flight, gateOccupationList)
                addition = True
                break

        return addition

    def handleNotFittingFlight(self, flight: Flight, gateOccupationList: dict, canceledFlights: list):
        """
        Handles the situation where a flight does not fit into any available gates or parking space.

        Parameters:
            flight (Flight): The flight that cannot be scheduled.
            gateOccupationList (dict): dictionary with the airport gate occupation data.
            canceledFlights (list): List with all the canceled flights.

        If no fitting space is found, attempts to reallocate the flight or cancels it.
        """
        if self.getFlightsCount(None, flight.arrival, flight.departure, gateOccupationList) < self.facts.airport.freeParkCount:
            if flight.aircraft.freePark:
                self.addOccupation(None, flight, gateOccupationList)
            else:
                if not self.reallocateFlight(flight, gateOccupationList):
                    canceledFlights.append(flight)

        else:
            canceledFlights.append(flight)
