# -*- coding: utf-8 -*-
"""
Znalostní systém - Modul rozhodování (riedl_inference)
Autor: Martin Riedl
Email: marty69@students.zcu.cz
Datum: 08.12.2024
Verze: 1.0

Popis:
Tento modul obsahuje implementaci inferenčního enginu, který rozhoduje, zda lze 
připravit konkrétní recept na základě aktuálních zásob a dostupných substitucí. 

Funkce modulu:
- Třída `InferenceEngine`: Obsahuje logiku pro ověření dostupnosti ingrediencí a rozhodování 
  o možnosti přípravy receptu.

Použití:
- Inicializace třídy `InferenceEngine` vyžaduje zásoby (`inventory`), recepty (`recipes`) 
  a tutoriály (`recipe_tutorials`).
- Metoda `can_cook(recipe_name)` určuje, zda je možné připravit zadaný recept.
- Podpůrná metoda `is_ingredient_or_sub_available` kontroluje dostupnost ingredience nebo její substituce.

Verze:
- 1.0: Základní implementace rozhodovací logiky.
"""

class InferenceEngine:
    """
    Třída pro rozhodování o možnosti přípravy receptu na základě dostupných ingrediencí.

    Atributy:
    - inventory: Instance pro správu zásob.
    - recipes: Instance pro správu receptů.
    - recipe_tutorials: Instance pro tutoriály receptů.

    Metody:
    - can_cook(recipe_name): Kontroluje, zda je možné připravit zadaný recept.
    - is_ingredient_or_sub_available(name, needed, available, subs): Kontroluje dostupnost ingredience 
      nebo její substituce.
    """
    def __init__(self, inventory, recipes, recipe_tutorials):
        """
        Inicializuje třídu `InferenceEngine`.

        Parametry:
        - inventory: Instance správy zásob.
        - recipes: Instance správy receptů.
        - recipe_tutorials: Instance pro tutoriály receptů.
        """
        self.inventory = inventory
        self.recipes = recipes
        self.recipe_tutorials = recipe_tutorials
    
    def can_cook(self, recipe_name):
        """
        Ověřuje, zda je možné připravit zadaný recept na základě dostupných zásob.

        Parametry:
        - recipe_name (str): Název receptu.

        Návratová hodnota:
        - tuple: (bool, list) - 
          True, pokud lze recept připravit, False jinak.
          List obsahuje chybějící ingredience ve formátu [(název, množství)] 
          nebo [(název, množství, 'nepodstatne')].
        """
        recipe = self.recipes.get_recipe(recipe_name)
        if not recipe:
            return False, []
        
        available = self.inventory.get_available_ingredients()
        missing = []
        
        for ingr in recipe['ingredience']:
            needed = ingr['mnozstvi']
            name = ingr['nazev']
            podstatne = ingr['podstatne']
            substituce = ingr['substituce']
            
            if self.is_ingredient_or_sub_available(name, needed, available, substituce):
                continue
            else:
                if podstatne:
                    missing.append((name, needed))
                else:
                    missing.append((name, needed, 'nepodstatne'))
        
        podstatne_chybi = [m for m in missing if len(m) == 2]
        if len(podstatne_chybi) > 0:
            return False, missing
        else:
            return True, missing
    
    def is_ingredient_or_sub_available(self, name, needed, available, subs):
        """
        Kontroluje, zda je požadovaná ingredience nebo její substituce dostupná v potřebném množství.

        Parametry:
        - name (str): Název ingredience.
        - needed (float): Požadované množství ingredience.
        - available (dict): Slovník dostupných ingrediencí.
        - subs (list): Seznam substitucí.

        Návratová hodnota:
        - bool: True, pokud je ingredience nebo některá substituce dostupná, False jinak.
        """
        if name in available and available[name]['mnozstvi'] >= needed:
            return True
        for s in subs:
            if s in available and available[s]['mnozstvi'] >= needed:
                return True
        return False
