# -*- coding: utf-8 -*-
"""
Znalostní systém - Modul správy zásob (riedl_inventory)
Autor: Martin Riedl
Email: marty69@students.zcu.cz
Datum: 13.12.2024
Verze: 2.2

Popis:
Tento modul spravuje skladové zásoby ingrediencí pro znalostní systém na vaření. 
Obsahuje funkce pro kontrolu, převod jednotek, aktualizaci množství a odečítání surovin 
z různých skladovacích prostor (lednice, mrazák, spíž, koření, nápoje).

Funkce modulu:
- Třída `Inventory`: Obsahuje logiku pro správu zásob, včetně kontroly dostupnosti ingrediencí, 
  převodu jednotek a odečítání surovin během vaření.

Použití:
- Inicializace třídy `Inventory` načte skladové zásoby z externích zdrojů.
- Metody umožňují kontrolu a manipulaci se zásobami podle potřeb uživatele.

Importované moduly:
- riedl_data_base.inventory.*: Obsahuje data o skladových zásobách (lednice, mrazák, atd.).

Verze:
- 2.0: Přidány funkce pro jednotkový převod a spotřebu ingrediencí.
- 2.2: Vylepšeno zpracování substitucí a chybové zprávy při nedostatku surovin.
"""

from .riedl_data_base.inventory.fridge import fridge
from .riedl_data_base.inventory.freezer import freezer
from .riedl_data_base.inventory.pantry import pantry
from .riedl_data_base.inventory.spices import spices
from .riedl_data_base.inventory.beverages import beverages

class Inventory:
    """
    Třída pro správu skladových zásob ingrediencí.

    Atributy:
    - UNIT_CONVERSIONS (dict): Definuje převody mezi jednotkami (g, kg, ml, l, ks).
    - fridge, freezer, pantry, spices, beverages: Skladové prostory s ingrediencemi.

    Metody:
    - get_available_ingredients(): Vrací kombinovaný seznam všech dostupných ingrediencí.
    - convert_units(amount, from_unit, to_unit): Převádí množství mezi různými jednotkami.
    - has_enough_ingredient(ingredient_name, needed_amount, needed_unit): Kontroluje, zda je k dispozici dostatek ingredience.
    - update_quantity(ingredient, amount): Aktualizuje množství ingredience v příslušném skladu.
    - consume_ingredients(ingredients_list, substitutions_used=None): Odečítá ingredience nebo substituce při vaření.
    """
    UNIT_CONVERSIONS = {
        'g': {'g': 1, 'kg': 1000},
        'kg': {'kg': 1, 'g': 0.001},
        'ml': {'ml': 1, 'l': 1000},
        'l': {'l': 1, 'ml': 0.001},
        'ks': {'ks': 1}  # Kusy nemají převod
    }

    def __init__(self):
        self.fridge = fridge
        self.freezer = freezer
        self.pantry = pantry
        self.spices = spices
        self.beverages = beverages

    def get_available_ingredients(self):
        """
        Vrací kombinovaný seznam všech dostupných ingrediencí z různých skladovacích prostor.

        Návratová hodnota:
        - dict: Slovník všech dostupných ingrediencí a jejich vlastností.
        """
        combined = {}
        for storage in [self.fridge, self.freezer, self.pantry, self.spices, self.beverages]:
            combined.update(storage)
        return combined

    def convert_units(self, amount, from_unit, to_unit):
        """
        Převádí množství z jedné jednotky na jinou.

        Parametry:
        - amount (float): Množství k převodu.
        - from_unit (str): Původní jednotka (např. 'g').
        - to_unit (str): Cílová jednotka (např. 'kg').

        Návratová hodnota:
        - float: Převáděné množství v cílové jednotce.

        Výjimky:
        - ValueError: Pokud není převod mezi jednotkami možný.
        """
        if from_unit == to_unit:
            return amount
        if from_unit in self.UNIT_CONVERSIONS and to_unit in self.UNIT_CONVERSIONS[from_unit]:
            conversion_rate = self.UNIT_CONVERSIONS[from_unit][to_unit]
            return amount * conversion_rate
        raise ValueError(f"Nelze převést z jednotky '{from_unit}' na '{to_unit}'.")

    def has_enough_ingredient(self, ingredient_name, needed_amount, needed_unit):
        """
        Kontroluje, zda je k dispozici dostatek ingredience.

        Parametry:
        - ingredient_name (str): Název ingredience.
        - needed_amount (float): Požadované množství.
        - needed_unit (str): Jednotka požadovaného množství.

        Návratová hodnota:
        - tuple: (bool, float) - True, pokud je dostatek; False jinak. Druhý prvek je dostupné množství.
        """
        all_ingr = self.get_available_ingredients()
        if ingredient_name not in all_ingr:
            return False, 0  # Ingredience neexistuje
        
        available = 0
        for storage in [self.fridge, self.freezer, self.pantry, self.spices, self.beverages]:
            if ingredient_name in storage:
                storage_amount = storage[ingredient_name]['mnozstvi']
                storage_unit = storage[ingredient_name]['jednotky']
                converted_amount = self.convert_units(storage_amount, storage_unit, needed_unit)
                available += converted_amount
        
        return available >= needed_amount, available
    
    def update_quantity(self, ingredient, amount):
        """
        Aktualizuje množství konkrétní ingredience v příslušném skladovacím prostoru.

        Parametry:
        - ingredient (str): Název ingredience.
        - amount (float): Množství k aktualizaci (kladné nebo záporné).

        Výjimky:
        - ValueError: Pokud je množství ingredience po aktualizaci záporné nebo pokud ingredience neexistuje.
        """
        location_map = {
            'fridge': self.fridge,
            'freezer': self.freezer,
            'pantry': self.pantry,
            'spices': self.spices,
            'beverages': self.beverages
        }

        ingredient_found = False

        for location, storage in location_map.items():
            if ingredient in storage:
                storage[ingredient]['mnozstvi'] += amount
                ingredient_found = True

                if storage[ingredient]['mnozstvi'] < 0:
                    raise ValueError(
                        f"Insufficient quantity of '{ingredient}' in {location}. "
                        f"Available: {storage[ingredient]['mnozstvi']}."
                    )
                
                print(f"DEBUG: Updated '{ingredient}' in {location}. New quantity: {storage[ingredient]['mnozstvi']}.")

        if not ingredient_found:
            print(f"DEBUG: Ingredient '{ingredient}' does not exist in any location.")
            raise ValueError(f"Ingredient '{ingredient}' does not exist in any location.")

    def consume_ingredients(self, ingredients_list, substitutions_used=None):
        """
        Odečítá ingredience nebo jejich substituce ze skladů při přípravě receptu.

        Parametry:
        - ingredients_list (list): Seznam požadovaných ingrediencí s množstvím a jednotkami.
        - substitutions_used (dict): Slovník použitých substitucí (volitelný).

        Výjimky:
        - ValueError: Pokud není dostatek ingrediencí pro přípravu receptu.
        """
        if substitutions_used is None:
            substitutions_used = {}
        
        for ingr in ingredients_list:
            needed = ingr['mnozstvi']
            needed_unit = ingr['jednotka']
            name = ingr['nazev']

            # Zkontrolovat hlavní ingredienci
            enough, available = self.has_enough_ingredient(name, needed, needed_unit)
            if not enough and 'substituce' in ingr and ingr['substituce']:
                # Hledat substituci
                for substitution in ingr['substituce']:
                    enough_sub, available_sub = self.has_enough_ingredient(substitution, needed, needed_unit)
                    if enough_sub:
                        name = substitution  # Použít substituci
                        substitutions_used[ingr['nazev']] = substitution
                        enough = True
                        break
            
            if not enough:
                raise ValueError(f"Nedostatek '{ingr['nazev']}'. Potřeba: {needed}{needed_unit}, dostupné: {available}{needed_unit}.")

            # Odečíst ingredienci nebo substituci
            for storage in [self.fridge, self.freezer, self.pantry, self.spices, self.beverages]:
                if name in storage:
                    storage_amount = storage[name]['mnozstvi']
                    storage_unit = storage[name]['jednotky']
                    to_deduct = self.convert_units(needed, needed_unit, storage_unit)
                    
                    if storage_amount >= to_deduct:
                        storage[name]['mnozstvi'] -= to_deduct
                        needed = 0
                        break
                    else:
                        needed -= self.convert_units(storage_amount, storage_unit, needed_unit)
                        storage[name]['mnozstvi'] = 0

            if needed > 0:
                raise ValueError(f"Nepodařilo se odečíst celé množství '{name}'. Zbývá: {needed}{needed_unit}.")
