# -*- coding: utf-8 -*-
"""
Znalostní systém - Modul postupů receptů (riedl_recipe_tutorials)
Autor: Martin Riedl
Email: marty69@students.zcu.cz
Datum: 13.12.2024
Verze: 1.1

Popis:
Tento modul poskytuje funkcionalitu pro zobrazení kroků přípravy receptů ve formátu tutoriálů. 
Umožňuje uživatelům projít recepty krok za krokem v přehledné a snadno čitelné podobě.

Funkce modulu:
- Třída `RecipeTutorials`: Centralizuje a spravuje kroky receptů.
- Metoda `get_recommendation`: Zobrazí detailní postup přípravy vybraného pokrmu.

Použití:
- Inicializace třídy `RecipeTutorials` načte kroky receptů z externích zdrojů.
- Zobrazení kroků pomocí metody `get_recommendation(name)`.

Importované moduly:
- riedl_data_base.tutorials.*: Obsahuje kroky jednotlivých receptů.
- riedl_misc.riedl_colors: Poskytuje funkce pro barevný výstup a formátování.

Verze:
- 1.0: Základní implementace pro zobrazení postupů receptů.
- 1.1: Přidáno barevné formátování a zlepšena čitelnost výstupu.
"""

from .riedl_data_base.tutorials.curry import steps as curry_recipe
from .riedl_data_base.tutorials.kung_pao import steps as kung_pao_recipe
from .riedl_data_base.tutorials.bramboracky import steps as bramboracky_recipe
from .riedl_data_base.tutorials.svickova import steps as svickova_recipe
from .riedl_data_base.tutorials.vepro_knedlo_zelo import steps as vepro_knedlo_zelo_recipe

from .riedl_misc.riedl_colors import  print_colored, print_header

class RecipeTutorials:
    """
    Třída pro správu kroků receptů a jejich zobrazení jako tutoriálů.

    Atributy:
    - recipe_dict (dict): Slovník receptů, kde klíčem je název receptu 
      a hodnotou jsou kroky přípravy.

    Metody:
    - get_recommendation(name): Zobrazí kroky přípravy vybraného receptu.
    """
    def __init__(self):
        """
        Inicializuje třídu `RecipeTutorials` a načte kroky dostupných receptů.
        """
        self.recipe_dict = {
            'kung_pao': kung_pao_recipe,          
            'curry': curry_recipe,                
            'bramboracky': bramboracky_recipe,    
            'svickova': svickova_recipe,          
            'vepro_knedlo_zelo': vepro_knedlo_zelo_recipe  
        }

    def get_recommendation(self, name):
        """
        Zobrazí krok za krokem recept v hezkém formátu.

        Parametry:
        - name (str): Název pokrmu, který uživatel chce připravit.

        Výstup:
        - Vypisuje kroky přípravy na konzoli.
        - Pokud recept neexistuje, vypíše chybovou zprávu.
        """
        steps = self.recipe_dict.get(name)
        if not steps:
            print_colored("Recept nebyl nalezen. Prosím, zkuste jiný pokrm.", "red")
            return

        # Vykreslení tutoriálu s formátováním
        print_header(f"Recept na {name.replace('_', ' ').capitalize()}")
        for step in steps:
            print_colored(step, "blue")
        print()  # Přidání prázdného řádku pro lepší čitelnost
