# -*- coding: utf-8 -*-
"""
Znalostní systém - Modul doporučování obchodů (riedl_store)
Autor: Martin Riedl
Email: marty69@students.zcu.cz
Datum: 08.12.2024
Verze: 1.0

Popis:
Tento modul obsahuje implementaci tříd `Store` a `StoreRecommender` pro práci s obchody a doporučování obchodů 
na základě dostupnosti ingrediencí, ceny, vzdálenosti a otevírací doby.

Funkce modulu:
- Třída `Store`: Reprezentuje jednotlivý obchod s vlastnostmi jako sortiment, otevírací doba a vzdálenost.
- Třída `StoreRecommender`: Obsahuje logiku pro doporučování obchodů s ohledem na dostupnost ingrediencí 
  a aktuální otevírací dobu.

Použití:
- Instance třídy `StoreRecommender` umožňuje vyhledávat obchody, kde lze zakoupit požadované ingredience, 
  a doporučuje nejvhodnější obchody spolu s důvody.

Importované moduly:
- math: Poskytuje matematické funkce pro výpočty.
- riedl_data_base.store.*: Obsahuje informace o jednotlivých obchodech (Tesco, Lidl, Albert, Rohlik, Wolt).
"""

import math
from .riedl_data_base.store.tesco import store as tesco_store
from .riedl_data_base.store.lidl import store as lidl_store
from .riedl_data_base.store.albert import store as albert_store
from .riedl_data_base.store.rohlik import store as rohlik_store
from .riedl_data_base.store.wolt import store as wolt_store


class Store:
    """
    Reprezentuje obchod s informacemi o sortimentu, otevírací době a vzdálenosti.
    
    Atributy:
    - name (str): Název obchodu.
    - assortment (dict): Sortiment obchodu s informacemi o ingrediencích.
    - open_from (tuple): Otevírací doba (hodina, minuta).
    - open_to (tuple): Zavírací doba (hodina, minuta).
    - distance (float): Vzdálenost obchodu v kilometrech.

    Metody:
    - is_open(current_hour, current_minute): Kontroluje, zda je obchod otevřený v daný čas.
    - buy_ingredient(ingredient, amount): Pokusí se zakoupit určitou ingredienci v požadovaném množství.
    """
    def __init__(self, name, assortment, open_from, open_to, distance):
        self.name = name
        self.assortment = assortment
        self.open_from = open_from
        self.open_to = open_to
        self.distance = distance

    def is_open(self, current_hour, current_minute):
        """
        Kontroluje, zda je obchod otevřený v zadaný čas.

        Parametry:
        - current_hour (int): Aktuální hodina (0-23).
        - current_minute (int): Aktuální minuta (0-59).

        Návratová hodnota:
        - bool: True, pokud je obchod otevřený; False jinak.
        """
        oh, om = self.open_from
        ch, cm = self.open_to
        current = current_hour * 60 + current_minute
        from_m = oh * 60 + om
        to_m = ch * 60 + cm
        return from_m <= current <= to_m

    def buy_ingredient(self, ingredient, amount):
        """
        Pokusí se zakoupit ingredienci v požadovaném množství.

        Parametry:
        - ingredient (str): Název ingredience.
        - amount (int): Požadované množství ingredience.

        Návratová hodnota:
        - bool: True, pokud bylo zakoupeno dostatečné množství; False jinak.
        """
        if ingredient in self.assortment:
            mnozstvi = self.assortment[ingredient]['mnozstvi']
            na_sklade = self.assortment[ingredient]['na_sklade']
            total_available = mnozstvi * na_sklade

            if total_available >= amount:
                # Calculate the number of packages needed
                packages_needed = math.ceil(amount / mnozstvi)

                if na_sklade >= packages_needed:
                    self.assortment[ingredient]['na_sklade'] -= packages_needed
                    return True
                else:
                    # Not enough packages to fulfill the amount
                    return False
        return False


class StoreRecommender:
    """
    Obsahuje logiku pro doporučování obchodů na základě ingrediencí a aktuální otevírací doby.
    
    Atributy:
    - stores (list): Seznam instancí třídy `Store`, reprezentujících dostupné obchody.

    Metody:
    - recommend_stores_with_reasons(ingredient, amount_needed, hour, minute):
        Doporučí obchody, kde lze zakoupit požadovanou ingredienci.
    - get_store_by_name(name):
        Vrací instanci obchodu na základě jeho názvu.
    """
    def __init__(self):
        self.stores = [
            Store(**tesco_store),
            Store(**lidl_store),
            Store(**albert_store),
            Store(**rohlik_store),
            Store(**wolt_store)
        ]
    
    def recommend_stores_with_reasons(self, ingredient, amount_needed, hour, minute):
        """
        Doporučuje obchody na základě dostupnosti ingrediencí a otevírací doby.

        Parametry:
        - ingredient (str): Název požadované ingredience.
        - amount_needed (int): Požadované množství ingredience.
        - hour (int): Aktuální hodina (0-23).
        - minute (int): Aktuální minuta (0-59).

        Návratová hodnota:
        - list: Seznam doporučených obchodů ve formátu (název, cena, vzdálenost, důvod).
        """
        results = []
        for st in self.stores:
            if ingredient not in st.assortment:
                results.append((st.name, None, None, "Není v nabídce"))
            else:
                if not st.is_open(hour, minute):
                    price = st.assortment[ingredient].get("price", None)
                    results.append((st.name, price, st.distance, "Zavřeno"))
                else:
                    mnozstvi = st.assortment[ingredient]['mnozstvi']
                    na_sklade = st.assortment[ingredient]['na_sklade']
                    total_available = mnozstvi * na_sklade

                    if total_available < amount_needed:
                        results.append((st.name, st.assortment[ingredient].get("price", None), st.distance, "Není skladem"))
                    else:
                        results.append((st.name, st.assortment[ingredient]["price"], st.distance, None))
        # Debugging output (can be uncommented if needed)
        # print("DEBUG: recommend_stores_with_reasons results =", results)
        return results
    
    def get_store_by_name(self, name):
        """
        Vyhledává obchod podle názvu.

        Parametry:
        - name (str): Název obchodu (case-insensitive).

        Návratová hodnota:
        - Store: Instance třídy `Store`, pokud je nalezena; None jinak.
        """
        for store in self.stores:
            if store.name.lower() == name.lower():
                return store
        return None