# -*- coding: utf-8 -*-
"""
Znalostní systém - Modul komunikace s uživatelem (riedl_user_interface)
Autor: Martin Riedl
Email: marty69@students.zcu.cz
Datum: 08.12.2024
Verze: 1.0

Popis:
Tento modul implementuje uživatelské rozhraní pro interakci s uživatelem v rámci znalostního systému na vaření.
Poskytuje funkce pro:
- Zobrazení hlavního menu s možnostmi výběru.
- Správu aktuálního času.
- Kontrolu dostupnosti surovin pro přípravu receptů.
- Doplnění zásob a nákup chybějících ingrediencí.
- Zobrazení návodů na přípravu receptů.
- Odečítání spotřebovaných ingrediencí.

Použití:
- Třída `UserInterface` je inicializována potřebnými moduly (inventář, inference engine, recommender, recepty a tutoriály).
- Funkce `main_menu()` spustí hlavní menu a umožní uživateli interagovat s ostatními moduly.

Moduly:
- riedl_misc.riedl_colors: Definuje barvy a funkce pro barevný výstup v konzoli.
- riedl_inventory: Poskytuje informace o dostupných surovinách v inventáři.
- riedl_inference: Kontroluje, zda je možné připravit recept na základě dostupných surovin.
- riedl_recipes: Poskytuje informace o receptech.
- riedl_store: Doporučuje obchody pro dokoupení ingrediencí.
- riedl_recipe_tutorials: Poskytuje návody na přípravu receptů.
"""

from modules.riedl_misc.riedl_colors import (
    RED, GREEN, YELLOW, BLUE, CYAN, RESET,
    print_colored, print_header, print_table, print_subheader
)


class UserInterface:
    """
    Třída reprezentuje uživatelské rozhraní pro interakci se znalostním systémem.
    
    Atributy:
    - inventory: Instance správy inventáře ingrediencí.
    - inference: Instance inferenčního enginu pro doporučení receptů.
    - recommender: Instance doporučovacího systému pro nákup ingrediencí.
    - recipes: Instance správy receptů.
    - recipe_tutorials: Instance poskytující návody na přípravu receptů.
    - current_hour: Aktuálně nastavená hodina (pro doporučení obchodů).
    - current_minute: Aktuálně nastavená minuta (pro doporučení obchodů).

    Metody:
    - main_menu: Hlavní menu pro uživatele.
    - set_current_time: Nastavení aktuálního času pro doporučení obchodů.
    - check_dish: Kontrola, zda je možné připravit konkrétní pokrm.
    - handle_substitutions: Řešení substitucí chybějících surovin.
    - ask_after_missing: Interakce s uživatelem ohledně chybějících surovin.
    - buy_missing_ingredient: Proces nákupu konkrétní chybějící ingredience.
    - choose_store_to_buy: Výběr obchodu pro nákup surovin.
    - finish_cooking: Dokončení procesu vaření, odečtení spotřebovaných surovin.
    - show_inventory: Zobrazení aktuálních zásob ingrediencí.
    - buy_ingredient_manually: Manuální přidání ingrediencí do zásob.
    """
    
    def __init__(self, inventory, inference, recommender, recipe_tutorials, recipes):
        """
        Inicializuje třídu UserInterface s požadovanými moduly.
        """
        self.inventory = inventory
        self.inference = inference
        self.recommender = recommender
        self.recipes = recipes
        self.recipe_tutorials = recipe_tutorials
        self.current_hour = None
        self.current_minute = None

    def main_menu(self):
        """
        Spouští hlavní menu uživatelského rozhraní.
        Nabízí možnosti interakce s inventářem, recepty a doporučeními.
        """
        while True:
            print_header("Znalostní systém pro vaření")

            # Dynamické vypsání pokrmů
            dishes = list(self.recipes.recipe_dict.keys())
            for i, dish in enumerate(dishes, 1):
                print(f"{i}) Uvařit {dish}")
            base_idx = len(dishes) + 1
            print(f"{base_idx}) Zobrazit dostupné ingredience")
            print(f"{base_idx + 1}) Dokoupit ingredienci (nákup)")
            print(f"{base_idx + 2}) Konec")
            print_subheader("Vyberte akci")

            choice = input(YELLOW + "Vyberte akci: " + RESET)

            try:
                c = int(choice)
                if 1 <= c <= len(dishes):
                    self.set_current_time()
                    self.check_dish(dishes[c - 1])
                elif c == len(dishes) + 1:
                    self.show_inventory()
                elif c == len(dishes) + 2:
                    self.set_current_time()
                    self.buy_ingredient_manually()
                elif c == len(dishes) + 3:
                    print_colored("Ukončuji program.", "green")
                    break
                else:
                    print_colored("Neplatná volba, zkuste to znovu.", "red")
            except ValueError:
                print_colored("Neplatná volba, zkuste to znovu.", "red")

    def set_current_time(self):
        """
        Umožňuje uživateli nastavit aktuální čas, který je použit pro doporučení obchodů.
        """
        while True:
            print_header("Nastavení Času")
            try:
                h = int(input(YELLOW + "Zadejte hodinu (0-23): " + RESET))
                m = int(input(YELLOW + "Zadejte minutu (0-59): " + RESET))
                if 0 <= h <= 23 and 0 <= m <= 59:
                    self.current_hour = h
                    self.current_minute = m
                    print_colored(f"Čas nastaven na {h:02d}:{m:02d}.", "green")
                    return
                else:
                    print_colored("Neplatný čas, zkuste to znovu.", "red")
            except ValueError:
                print_colored("Neplatný vstup, zkuste to znovu.", "red")

    def check_dish(self, dish_name):
        """
        Kontroluje, zda je možné připravit konkrétní pokrm na základě dostupných ingrediencí.
        
        Parametry:
        - dish_name (str): Název receptu k přípravě.
        """
        can_cook, missing = self.inference.can_cook(dish_name)
        if can_cook and len(missing) == 0:
            print_colored(f"\nMůžete připravit '{dish_name}' s dostupnými ingrediencemi!", "green")
            # Display the cooking tutorial
            self.recipe_tutorials.get_recommendation(dish_name)
            self.finish_cooking(dish_name, {})
            return
        substitutions_used = {}
        new_missing = self.handle_substitutions(dish_name, missing, substitutions_used)

        if len(new_missing) == 0:
            print_colored(f"\nNyní můžete připravit '{dish_name}' s použitím substitucí!", "green")
            self.finish_cooking(dish_name, substitutions_used)
            return

        print_colored(f"\nBohužel, nemůžete ihned připravit '{dish_name}'. Chybí:", "red")
        for ingr in new_missing:
            print_colored(f"- {ingr[0]} (podstatné)", "red")
        self.ask_after_missing(dish_name, new_missing, substitutions_used)

    def handle_substitutions(self, dish_name, missing, substitutions_used):
        """
        Zpracovává substituce pro chybějící ingredience receptu.
        
        Parametry:
        - dish_name (str): Název pokrmu, který se uživatel pokouší připravit.
        - missing (list): Seznam chybějících ingrediencí ve formátu [(ingredience, množství)].
        - substitutions_used (dict): Slovník použitých substitucí ve formátu {originál: substituce}.
        
        Návratová hodnota:
        - new_missing (list): Seznam ingrediencí, které stále chybí i po zohlednění substitucí.
        """
        recipe = self.recipes.get_recipe(dish_name)
        if not recipe:
            return missing
        available = self.inventory.get_available_ingredients()
        new_missing = []
        for ingr in missing:
            name = ingr[0]
            needed = ingr[1]
            recipe_ingr = None
            for ri in recipe['ingredience']:
                if ri['nazev'] == name:
                    recipe_ingr = ri
                    break
            if not recipe_ingr:
                new_missing.append(ingr)
                continue
            subs = recipe_ingr.get('substituce', [])
            used_sub = None
            if subs:
                for s in subs:
                    if s in available and available[s]['mnozstvi'] >= needed:
                        used_sub = s
                        break
            if used_sub:
                substitutions_used[name] = used_sub
            else:
                new_missing.append(ingr)
        return new_missing

    def ask_after_missing(self, dish_name, missing, substitutions_used):
        """
        Umožňuje uživateli rozhodnout, zda chce dokoupit chybějící ingredience nebo jejich substituce.
        
        Parametry:
        - dish_name (str): Název pokrmu, který se uživatel pokouší připravit.
        - missing (list): Seznam chybějících ingrediencí.
        - substitutions_used (dict): Slovník použitých substitucí.
        """
        print_header("Dokoupit Chybějící Suroviny")
        print_colored("Chcete se pokusit dokoupit některou z chybějících surovin nebo jejich substituci?", "yellow")
        print("1) " + GREEN + "Ano" + RESET)
        print("2) " + RED + "Ne, vrátit se do hlavního menu" + RESET)

        choice = input(YELLOW + "Vyberte (1-2): " + RESET)

        if choice == '1':
            print(YELLOW + "\nChybějící suroviny (včetně substitucí, pokud byly použity):" + RESET)
            for idx, ingr in enumerate(missing, 1):
                ingr_name = ingr[0]
                if ingr_name in substitutions_used:
                    substitution = substitutions_used[ingr_name]
                    print_colored(f"{idx}) {ingr_name} (nahrazeno: {substitution})", "yellow")
                else:
                    print_colored(f"{idx}) {ingr_name}", "yellow")
            idx_str = input(YELLOW + "Vyberte číslo ingredience k dokoupení: " + RESET)
            try:
                idx = int(idx_str) - 1
                if 0 <= idx < len(missing):
                    selected_ingr = missing[idx]
                    self.buy_missing_ingredient(selected_ingr)
                    can_cook, new_missing = self.inference.can_cook(dish_name)
                    new_missing = self.handle_substitutions(dish_name, new_missing, substitutions_used)
                    if len(new_missing) == 0 and can_cook:
                        print_colored(f"Nyní už můžete připravit {dish_name}!", "green")
                        # Display the cooking tutorial
                        self.recipe_tutorials.get_recommendation(dish_name)
                        self.finish_cooking(dish_name, substitutions_used)
                    else:
                        print_colored("Stále chybí některé ingredience.", "red")
                        self.ask_after_missing(dish_name, new_missing, substitutions_used)
                else:
                    print_colored("Neplatná volba.", "red")
                    self.ask_after_missing(dish_name, missing, substitutions_used)
            except ValueError:
                print_colored("Neplatná volba. Zadejte číslo.", "red")
                self.ask_after_missing(dish_name, missing, substitutions_used)
        elif choice == '2':
            print_colored("Vracíme se do hlavního menu.", "red")
        else:
            print_colored("Neplatná volba, vracím se do hlavního menu.", "red")

    def buy_missing_ingredient(self, ingr_info):
        """
        Zajišťuje nákup konkrétní ingredience, která chybí.
        
        Parametry:
        - ingr_info (tuple): Informace o ingredienci ve formátu (název, množství).
        
        Popis:
        - Doporučuje obchody, kde lze ingredienci zakoupit.
        - Aktualizuje inventář po úspěšném nákupu.
        """
        name = ingr_info[0]
        amount_needed = ingr_info[1]

        # Získání informací o obchodech
        stores_info = self.recommender.recommend_stores_with_reasons(name, amount_needed, self.current_hour, self.current_minute)
        print_header(f"Surovina: '{name}'")

        if not stores_info:
            print_colored("Žádné dostupné obchody pro tento nákup.\n", "red")
            return

        # Příprava dat pro tabulku
        headers = ["Číslo", "Obchod", "Cena", "Vzdálenost (km)", "Status"]
        rows = []
        for i, info in enumerate(stores_info, 1):
            sname, price, dist, reason = info
            store = self.recommender.get_store_by_name(sname)
            if store and store.open_from == (0, 0) and store.open_to == (23, 59):
                # Speciální označení pro obchody otevřené 24/7
                if reason is None:
                    status = "Dostupné (24/7)"
                else:
                    status = f"Nedostupné (24/7) - {reason}"
            else:
                if reason is None:
                    status = "Dostupné"
                else:
                    status = f"Nedostupné - {reason}"
            price_display = f"{price} Kč"
            rows.append([str(i), sname, price_display, f"{dist} km", status])

        print_table(headers, rows)
        print()

        self.choose_store_to_buy(stores_info, name, amount_needed)

    def choose_store_to_buy(self, stores_info, name, amount_needed):
        """
        Umožňuje uživateli vybrat obchod pro nákup ingrediencí.
        
        Parametry:
        - stores_info (list): Seznam obchodů s dostupnými informacemi o ceně, vzdálenosti a dostupnosti.
        - name (str): Název ingredience k zakoupení.
        - amount_needed (int): Potřebné množství ingredience.
        """
        while True:
            choice_str = input(YELLOW + "Vyberte číslo obchodu, odkud chcete koupit (nebo 'q' pro zrušení): " + RESET)
            if choice_str.lower() == 'q':
                print_colored("Nákup zrušen.\n", "red")
                break
            try:
                choice = int(choice_str) - 1
                if 0 <= choice < len(stores_info):
                    sname, price, dist, reason = stores_info[choice]
                    if reason is not None:
                        print_colored(f"Nelze koupit z {sname}: {reason}\n", "red")
                        continue
                    store = self.recommender.get_store_by_name(sname)
                    if store and store.buy_ingredient(name, amount_needed):
                        self.inventory.update_quantity(name, amount_needed)
                        print_colored("\n==============================", "green")
                        print_colored("   Nákup úspěšný", "green")
                        print_colored("==============================", "green")
                        print_colored(f"Úspěšně zakoupeno množství '{amount_needed}' '{name}' v obchodě '{sname}' a přidáno do zásob!\n", "green")
                        break
                    else:
                        print_colored(f"Nákup v {sname} se nezdařil. Zkuste to prosím znovu.\n", "red")
                else:
                    print_colored("Neplatná volba. Zkuste to prosím znovu.\n", "red")
            except ValueError:
                print_colored("Neplatný vstup. Zadejte číslo odpovídající obchodu.\n", "red")

    def finish_cooking(self, dish_name, substitutions_used):
        """
        Dokončuje proces vaření pokrmu.
        
        Parametry:
        - dish_name (str): Název receptu, který byl připraven.
        - substitutions_used (dict): Slovník použitých substitucí během vaření.
        
        Popis:
        - Odečítá spotřebované suroviny z inventáře.
        - Zobrazuje potvrzení a informace o použitých substitucích.
        """
        recipe = self.recipes.get_recipe(dish_name)
        if recipe:
            self.inventory.consume_ingredients(recipe['ingredience'], substitutions_used=substitutions_used)
            print_colored(f"{dish_name} bylo úspěšně připraveno! Suroviny byly odečteny.", "green")
            if substitutions_used:
                print_colored("Použité substituce:", "yellow")
                for original, substitute in substitutions_used.items():
                    print_colored(f"- {original} nahrazeno {substitute}", "yellow")
        else:
            print_colored("Recept nenalezen, nelze odečíst suroviny.", "red")

    def show_inventory(self):
        """
        Zobrazuje aktuální zásoby ingrediencí uživatele.
        
        Popis:
        - Vypisuje tabulku obsahující název ingredience, množství a jednotky.
        - Zvýrazňuje ingredience s nulovým množstvím červeně.
        """
        combined = self.inventory.get_available_ingredients()
        print_header("Aktuální zásoby")

        if not combined:
            print_colored("Zásoby jsou prázdné.", "red")
            return

        # Příprava dat pro tabulku
        headers = ["Surovina", "Množství", "Jednotky"]
        rows = []
        for ingr, details in sorted(combined.items()):
            quantity = details.get('mnozstvi', 0)
            units = details.get('jednotky', '')
            quantity_display = f"{quantity}"
            if quantity == 0:
                quantity_display = RED + str(quantity) + RESET
            else:
                quantity_display = GREEN + str(quantity) + RESET
            rows.append([ingr, quantity_display, units])

        print_table(headers, rows)
        print()

    def buy_ingredient_manually(self):
        """
        Umožňuje uživateli ručně dokoupit ingredience a přidat je do inventáře.
        
        Popis:
        - Uživatel zadá název ingredience a její množství.
        - Ingredience je přidána do inventáře po úspěšném zadání.
        """
        while True:
            print_header("Dokoupit Ingredience")

            name = input(YELLOW + "Zadejte název ingredience k nákupu: " + RESET).strip()
            if not name:
                print_colored("Název ingredience nemůže být prázdný.", "red")
                continue

            amount_str = input(YELLOW + "Zadejte množství (číslo): " + RESET).strip()
            try:
                amount = int(amount_str)
                if amount <= 0:
                    print_colored("Množství musí být kladné číslo.", "red")
                    continue
                self.buy_missing_ingredient((name, amount))
                print_colored(f"Úspěšně zakoupeno množství '{amount}' '{name}'.", "green")
                break  # Úspěšný nákup, opustíme smyčku
            except ValueError:
                print_colored("Neplatné množství. Zadejte celé číslo.", "red")
