"""
Rules Module for Linear Equation Solver
Faiz Suleimanov
13.12.2024
version 1.0

This module defines the `Rule` base class and its specific implementations for manipulating
facts in the knowledge base. These rules enable adding, removing, replacing, and eliminating
variables in equations.
"""

from sympy import solve, simplify, Eq

class Rule:
    """
    Base class for all rules. A rule defines an action that can be applied to the knowledge base.
    """
    def apply(self, kb):
        pass

    def __str__(self):
        return self.__class__.__name__


class AddFactRule(Rule):
    """
    A rule to add a fact to the knowledge base.
    """
    def __init__(self, fact):
        self.fact = fact

    def apply(self, kb):
        """
        Add the fact to the knowledge base if it does not already exist.
        :param kb: The knowledge base instance.
        """
        if self.fact not in kb.facts:
            kb.add_fact(self.fact)
            print(f"Fact added: {self.fact}")

    def __str__(self):
        return f"AddFactRule(fact={self.fact})"


class RemoveFactRule(Rule):
    """
    A rule to remove a fact from the knowledge base.
    """
    def __init__(self, fact):
        self.fact = fact

    def apply(self, kb):
        """
        Remove the fact from the knowledge base if it exists.
        :param kb: The knowledge base instance.
        """
        if self.fact in kb.facts:
            kb.remove_fact(self.fact)
            print(f"Fact removed: {self.fact}")

    def __str__(self):
        return f"RemoveFactRule(fact={self.fact})"


class ReplaceFactRule(Rule):
    """
    A rule to replace an old fact with a new fact in the knowledge base.
    """
    def __init__(self, old_fact, new_fact):
        self.old_fact = old_fact
        self.new_fact = new_fact

    def apply(self, kb):
        """
        Replace the old fact with the new fact if the old fact exists in the knowledge base.
        :param kb: The knowledge base instance.
        """
        if self.old_fact in kb.facts:
            kb.remove_fact(self.old_fact)
            print(f"Fact removed: {self.old_fact}")
            if self.new_fact not in kb.facts:
                kb.add_fact(self.new_fact)
                print(f"Fact added: {self.new_fact}")

    def __str__(self):
        return f"ReplaceFactRule(old_fact={self.old_fact}, new_fact={self.new_fact})"


class EliminateVariableRule(Rule):
    """
    A rule to eliminate a variable from the knowledge base using substitution.
    """
    def __init__(self, variable):
        self.variable = variable

    def apply(self, kb):
        """
         Eliminate the specified variable by substituting its value in relevant equations.
         :param kb: The knowledge base instance.
         """
        print(f"Processing variable: {self.variable}")
        relevant_facts = kb.find_facts_with_variable(self.variable)

        if len(relevant_facts) < 1:
            print(f"Not enough equations to exclude variable {self.variable}")
            return

        eq1 = relevant_facts[0]
        solved_for_var = solve(eq1, self.variable)

        if solved_for_var:
            substitution = solved_for_var[0]
            print(f"Solution found for {self.variable}: {substitution}")

            kb.add_rule(RemoveFactRule(eq1))

            for fact in kb.facts[:]:
                if fact != eq1:
                    simplified_fact = simplify(fact.subs(self.variable, substitution))
                    if simplified_fact != True and simplified_fact != False:
                        kb.add_rule(ReplaceFactRule(fact, simplified_fact))

            kb.add_rule(AddFactRule(Eq(self.variable, substitution)))
            print(f"Rules for variable {self.variable} were added")

    def __str__(self):
        return f"EliminateVariableRule(variable={self.variable})"
