"""
User Interface Module for Linear Equation Solver
Faiz Suleimanov
13.12.2024
version 1.0

This module defines the `UserInterface` class, which provides a console-based interactive interface
for solving linear equations. It guides the user through variable input, equation input, and the
solution process while offering explanations for each step when requested.
"""

from modules.rules import EliminateVariableRule
from modules.solver import solve_gaus
from modules.data_manager import DataManager
from modules.guide import print_guide, explain_step
from colorama import init, Fore, Style


class UserInterface:
    """
    A class to provide a console-based user interface for interacting with the linear equation solver.
    """
    def __init__(self):
        self.data_manager = DataManager()
        init()

    def print_header(self, text):
        """
        Print a formatted header with the given text.
        :param text: The header text to display.
        """
        print(f"\n{Fore.CYAN}" + "=" * 50 + Style.RESET_ALL)
        print(f"{Fore.CYAN}{text:^50}{Style.RESET_ALL}")
        print(f"{Fore.CYAN}" + "=" * 50 + f"{Style.RESET_ALL}\n")

    def print_step(self, step_num, text):
        """
        Print a formatted step with its number and description.
        :param step_num: The step number.
        :param text: The description of the step.
        """
        print(f"\n{Fore.GREEN}[Step {step_num}]{Style.RESET_ALL} {Fore.YELLOW}{text}{Style.RESET_ALL}")
        print(f"{Fore.GREEN}" + "-" * 50 + f"{Style.RESET_ALL}")

    def get_variables(self):
        """
        Prompt the user to input variables for the equations.
        :return: True if variables are successfully set, False otherwise.
        """
        self.print_header("VARIABLE INPUT")
        print(
            f"""{Fore.YELLOW}Enter variables separated by space (example: x y z w) 
             or 'explain' for getting an explanation.{Style.RESET_ALL}""")

        while True:
            print(f"{Fore.BLUE}> {Style.RESET_ALL}", end="")
            user_input = input().lower()

            if user_input == 'help':
                print_guide()
                print(f"\n{Fore.YELLOW}Now, enter your variables separated by space.{Style.RESET_ALL}")
                print(f"{Fore.BLUE}> {Style.RESET_ALL}", end="")
                continue
            if user_input == 'explain':
                explain_step("variables_input")
                print(f"\n{Fore.YELLOW}Now, enter your variables separated by space.{Style.RESET_ALL}")
                continue
            variable_names = user_input.split()
            if not variable_names:
                print(f"{Fore.RED}✖ Error: Please enter at least one variable{Style.RESET_ALL}")
                continue

            return self.data_manager.set_variables(variable_names)

    def input_equations(self):
        """
        Prompt the user to input equations for the system.
        :return: True if equations are successfully entered, False otherwise.
        """
        self.print_header("EQUATION INPUT")
        print(f"{Fore.YELLOW}Enter equations one by one. Type 'done' when finished{Style.RESET_ALL}")
        print(f"{Fore.YELLOW}Example: 7*x + 3*y - 2*z + 5*w = 18{Style.RESET_ALL}")
        print(f"{Fore.YELLOW}Type 'help' to see the guide or 'explain' for an explanation{Style.RESET_ALL}")

        equation_count = 1
        while True:
            print(f"\n{Fore.CYAN}Equation #{equation_count}{Style.RESET_ALL}")
            print(f"{Fore.BLUE}> {Style.RESET_ALL}", end="")
            equation_input = input().lower()

            if equation_input == 'help':
                print_guide()
                continue
            elif equation_input == 'explain':
                explain_step("equation_input")
                continue
            elif equation_input == 'done':
                if equation_count == 1:
                    print(
                        f"{Fore.RED}✖ Error: No equations were entered. At least one equation is required.{Style.RESET_ALL}")
                    return False
                break

            result = self.data_manager.add_equation(equation_input)
            if result is not True:
                print(f"{Fore.RED}✖ Error: {result}{Style.RESET_ALL}")
            else:
                print(f"{Fore.GREEN}✓ Equation added successfully{Style.RESET_ALL}")
                equation_count += 1
        return True

    def solve_equations(self):
        """
        Solve the system of equations by applying rules and Gaussian elimination.
        """
        kb = self.data_manager.get_knowledge_base()
        variables = self.data_manager.get_variables()

        self.print_header("SOLVING SYSTEM OF EQUATIONS")

        self.print_step(1, "Initial system of equations")
        print(f"{Fore.WHITE}{kb}{Style.RESET_ALL}")

        self.print_step(2, "Variable elimination process")
        for i, var in enumerate(variables, 1):
            print(f"\n{Fore.YELLOW}Eliminating variable '{var}':{Style.RESET_ALL}")
            kb.add_rule(EliminateVariableRule(var))
            print(f"\n{Fore.CYAN}Current system:{Style.RESET_ALL}")
            print(f"{Fore.WHITE}{kb}{Style.RESET_ALL}")

            if i < len(variables):
                print(
                    f"\n{Fore.YELLOW}Press Enter to continue or type 'explain' for an explanation...{Style.RESET_ALL}",
                    end="")
                print(f"\n{Fore.BLUE}> {Style.RESET_ALL}", end="")
                user_input = input().lower()
                if user_input == 'explain':
                    explain_step("variable_elimination", var)
            kb.apply_rules()
        input(f"{Fore.YELLOW}\nPres 'enter' for the continue.{Style.RESET_ALL}")
        self.print_step(3, "System after transformations")
        print(f"\n{Fore.WHITE}{kb}{Style.RESET_ALL}")

        input(f"{Fore.YELLOW}\nPres 'enter' for the solution.{Style.RESET_ALL}")
        self.print_step(4, "Final solution")
        print(f"{Fore.CYAN}Solving using Gaussian elimination:{Style.RESET_ALL}")
        solve_gaus(kb, variables)

        self.print_header("SOLUTION COMPLETED")

    def run(self):
        """
        Run the interactive user interface.
        """
        self.print_header("WELCOME TO LINEAR EQUATION SOLVER")
        print(f"{Fore.YELLOW}Type 'help' at any time to see the guide{Style.RESET_ALL}")

        while True:
            self.data_manager.reset()
            self.get_variables()
            if not self.input_equations():
                continue
            self.solve_equations()

            print(f"\n{Fore.YELLOW}Do you want to run the program again? (yes/no){Style.RESET_ALL} ", end="")
            print(f"\n{Fore.BLUE}> {Style.RESET_ALL}", end="")
            if input().lower() != 'yes':
                break
