from Data.GameState import GameState
from Data.Move import Move
from Data.ActionType import ActionType
from Knowledge.Knowledge import Knowledge
from Explanation.Explanation import Explanation
from Communication.CommunicationModule import CommunicationModule

import copy


class InferenceModule:
    """
    Inferencni modul pro hru NIM
    """

    def __init__(self):
        """
        Inicializuje znalosti - ktere stavy uz byly zpracovany
        """
        self.knowledge = Knowledge()

    def __get_moves(self, game_state: GameState) -> [Move]:
        """
        Vrati vsechny mozne tahy pro dany stav hry
        :param game_state: Stav hry (hromadky)
        :return: Vraci pole moznych tahu
        """
        moves = []

        # Pridani jednoduchych tahu
        for i in range(len(game_state.piles)):  # Prvni hromadka
            for j in range(1, game_state.piles[i] + 1):  # Od 1 do maximalniho poctu prvku v hromadce
                moves.append(Move(ActionType.SINGLE, j, [i]))

        # Pridani dvojitych tahu
        for i in range(len(game_state.piles)):  # Prvni hromadka
            for j in range(i + 1, len(game_state.piles)):  # Druha hromadka

                max_amount = min(game_state.piles[i], game_state.piles[j])  # Maximum, ktere obsahuji obe hromadky

                # Od 1 do maximalniho poctu prvku v obou hromadkach
                for k in range(1, max_amount + 1):
                    moves.append(Move(ActionType.DOUBLE, k, [i, j]))

        return moves

    def get_piles_occupied(self, game_state: GameState):
        """
        Metoda pocita pocet hromadek, ktere obsahuji alespon jednu zapalku
        :param game_state: Stav hry
        :return: Vraci pocet hromadek, ktere obsahuji alespon jednu zapalku
        """
        piles_occupied = 0

        for i in range(len(game_state.piles)):
            if game_state.piles[i] != 0:
                piles_occupied += 1

        return piles_occupied

    def game_ended(self, game_state: GameState):
        """
        Kontrola, jestli je hra v koncovem stavu (v zadne hromadce neni ani jedna zapalka)
        :param game_state: Stav hry
        :return: Vrati True, pokud je konec hry, jinak False
        """

        return self.get_piles_occupied(game_state) == 0

    def end_state(self, game_state: GameState):
        """
        Kontrola, jestli je hra v koncovem stavu (je to vyherni pozice pro jednoho z hracu)
        Jinymi slovy to je posledni kolo, kde jeden z hracu vyhraje
        :param game_state: Stav hry
        :return: True, pokud je konec hry, jinak false
        """
        piles_occupied = 0
        nim_sum = 0

        for i in range(len(game_state.piles)):
            if game_state.piles[i] != 0:
                piles_occupied += 1

            nim_sum ^= game_state.piles[i]

        # Pokud je pouze jedna hromadka
        if piles_occupied == 1:
            return True

        # Pokud jsou dve hromadky, ale se stejnym poctem prvku
        if piles_occupied == 2 and nim_sum == 0:
            return True

        return False

    def apply_move(self, game_state: GameState, move: Move):
        """
        Aplikuje tah na stav hry
        :param game_state: Stav hry
        :param move: Tah, ktery se ma aplikovat
        """
        # Pokud je vyber pouze z jedne hromadky
        if move.action_type == ActionType.SINGLE:
            game_state.piles[move.pile_numbers[0]] -= move.amount
            return

        # Pokud je vyber ze dvou hromadek
        game_state.piles[move.pile_numbers[0]] -= move.amount
        game_state.piles[move.pile_numbers[1]] -= move.amount

    def revert_move(self, game_state: GameState, move: Move):
        """
        Vraceni tahu zpet
        :param game_state: Stav hry
        :param move: Tah, ktery se ma vratit
        """
        # Inverzni tah a jeho nasledne pouziti na stav hry
        inverse_move = Move(move.action_type, -move.amount, move.pile_numbers)
        self.apply_move(game_state, inverse_move)

    def minmax(self, is_maximizing: bool, game_state: GameState) -> int:
        """
        Minimax algoritmus pro hru NIM
        :param is_maximizing: Priznak, jestli se jedna o maximalizujiciho hrace
        :param game_state: Stav hry
        :return: Skore, ktere ma maximalizujici hrac (1 pokud vyhraje, -1 pokud prohraje)
        """

        # Pokud je konec hry
        if self.end_state(game_state):
            return 1 if is_maximizing else -1

        moves = self.__get_moves(game_state)
        # Jedna se o maximalizujiciho hrace
        if is_maximizing:
            best_score = -2
            for current_move in moves:
                self.apply_move(game_state, current_move)

                score = 0
                if self.knowledge.contains_state(game_state):
                    score = self.knowledge.get_result(game_state, True)
                else:
                    score = self.minmax(not is_maximizing, game_state)
                    self.knowledge.add_state(copy.deepcopy(game_state), score, True)

                self.revert_move(game_state, current_move)

                best_score = max(score, best_score)

            return best_score

        # Jedna se o minimalizujiciho hrace
        best_score = 2
        for current_move in moves:
            self.apply_move(game_state, current_move)

            score = 0
            if self.knowledge.contains_state(game_state):
                score = self.knowledge.get_result(game_state, False)
            else:
                score = self.minmax(not is_maximizing, game_state)
                self.knowledge.add_state(copy.deepcopy(game_state), score, False)

            self.revert_move(game_state, current_move)

            best_score = min(score, best_score)

        return best_score

    def find_best_move(self, game_state: GameState) -> Move:
        """
        Vyhodnoti tah po tahu kazdy, zjisti, jestli vedou k vyhre a vybere ten nejlepsi
        :param game_state: Stav hry
        :return: Navrati nejlepsi tah
        """
        moves = self.__get_moves(game_state)

        best_score = -2
        best_move = None

        for move in moves:
            self.apply_move(game_state, move)

            score = 0

            if self.knowledge.contains_state(game_state):
                score = self.knowledge.get_result(game_state, True)
            else:
                score = self.minmax(False, game_state)
                self.knowledge.add_state(copy.deepcopy(game_state), score, True)

            self.revert_move(game_state, move)

            if score > best_score:
                best_score = score
                best_move = move

        return best_move

    def inference_game_init(self) -> GameState:
        """
        Spousti inferenci pro hru NIM
        :return: Vraci pocatecni stav hry
        """

        number_of_piles = CommunicationModule.numeric_input("pocet hromadek",
                                                            "Zadejte pocet hromadek, ze kterych chcete v ramci hry vybirat")
        piles = []

        for i in range(number_of_piles):
            piles.append(CommunicationModule.numeric_input(f"pocet zapalek pro {i + 1}. hromadku",
                                                           f"Zadejte pocet sirek v hromadce cislo {i + 1}"))

        return GameState(piles)

    def get_move_choices(self, game_state):
        """
        Vrati navrhovane moznosti (jestli lze odebrat z dvou hromadek, nebo jen z jedne)
        :param game_state: Stav hry
        :return: Vrati seznam moznosti
        """
        piles_occupied = self.get_piles_occupied(game_state)
        choices = ["Odebrat z jedne hromadky libovolny pocet zapalek"]

        if piles_occupied > 1:
            choices.append("Odebrat ze dvou hromadek stejne mnozstvi zapalek")
        choices.append("Vybrat automaticky optimalni krok")

        return choices

    def choose_pile(self, game_state, name):
        """
        Metoda ziskava cislo hromadky, ze ktere chce uzivatel odebrat zapalky
        :param game_state: Stav hry
        :param name: Jmeno hromadky (prvni, druha, ...)
        :return: Vraci cislo hromadky
        """
        while True:
            pile_number = CommunicationModule.numeric_input_bounded(name,
                                                                    f"Zadejte {name}, ze ktere chcete odebrat zapalky, {name}. Stav hry {game_state}",
                                                                    1, len(game_state.piles))
            if game_state.piles[pile_number - 1] == 0:
                Explanation.show_explanation("Hromadka je prazdna, vyberte jinou hromadku")
                continue
            return pile_number

    def inference_move(self, game_state: GameState) -> Move:
        """
        Methoda ziskava tah od uzivatele
        :param game_state: Stav hry
        :return: Vraci instanci reprezentujici zvoleny tah
        """

        choices = self.get_move_choices(game_state)

        while True:
            choice = CommunicationModule.user_choice("druh tahu", choices,
                                                     f"Vyberte jednu z moznosti pro nasledujici stav hry:\n{game_state}")

            if choice == 1:
                pile_number = self.choose_pile(game_state, "cislo hromadky")

                amount = CommunicationModule.numeric_input_bounded("pocet zapalek",
                                                                   f"Zadejte pocet zapalek, ktere chcete odebrat. Stav hry {game_state}",
                                                                   1,
                                                                   game_state.piles[pile_number - 1])

                return Move(ActionType.SINGLE, amount, [pile_number - 1])

            # Pokud jsou na vyber 3 moznosti
            if len(choices) == 3:
                if choice == 2:

                    pile_number1 = self.choose_pile(game_state, "cislo prvni hromadky")

                    pile_number2 = 0

                    # Zadani druhe hromadky, dokud se nezada jina nez prvni
                    while True:
                        pile_number2 = self.choose_pile(game_state, "cislo druhe hromadky")

                        if pile_number1 != pile_number2:
                            break

                        Explanation.show_explanation("Nelze vybrat stejne hromadky")

                    max_amount = min(game_state.piles[pile_number1 - 1], game_state.piles[pile_number2 - 1])
                    amount = CommunicationModule.numeric_input_bounded("pocet zapalek",
                                                                       f"Zadejte pocet zapalek, ktere chcete odebrat (maximalne {max_amount}). Stav hry {game_state}",
                                                                       1, max_amount)

                    return Move(ActionType.DOUBLE, amount, [pile_number1 - 1, pile_number2 - 1])

                if choice == 3:
                    return self.find_best_move(game_state)

            # Pokud jsou na vyber pouze 2 moznosti
            else:
                if choice == 2:
                    return self.find_best_move(game_state)

    def start_game(self):
        """
        Spousti hru NIM
        """
        game_state = self.inference_game_init()
        player1 = CommunicationModule.string_input("jmeno prvniho hrace")
        player2 = CommunicationModule.string_input("jmeno druheho hrace")

        players = [player1, player2]

        current_player = 0

        while not self.game_ended(game_state):
            Explanation.show_explanation(f"Na tahu je hrac {players[current_player]}")

            CommunicationModule.write_output(str(game_state))

            # Nacteni tahu
            move = self.inference_move(game_state)

            Explanation.show_explanation(f"Hrac {players[current_player]} provedl tah: {move}")
            self.apply_move(game_state, move)

            # Prepnuti na druheho hrace
            current_player = (current_player + 1) % len(players)

        CommunicationModule.write_output(
            f"\n\nVyhral hrac {players[(current_player + len(players) - 1) % len(players)]}\n\n")

    def infere(self):
        """
        Spousti inferenci pro hru NIM
        """

        while True:
            self.start_game()
            choice = CommunicationModule.user_choice("funkce aplikace", ["Pokracovat v dalsi hre", "Ukoncit aplikaci"],
                                                     f"Vyberte, jestli chcete pokracovat v hrani dalsi partie nebo ukoncit aplikaci")
            if choice == 2:
                break
